﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 * 
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion
using System;
using System.Diagnostics.Contracts;
using HeuristicLab.Random;

namespace HeuristicLab.Algorithms.DataAnalysis.MctsSymbolicRegression {

  // calculates a hash-code for expressions.
  // The hash codes of equivalent expressions x * (x + y) = (y + x)*x = (yx * xx) must be the same.
  // We use a random number for each symbol in the expressions and use symmetric operations to produce
  // results of sub-expressions.
  // Surrogate functions for the functions log, exp, 1/x are necessary.
  // Ideally, the surrogate functions also match identities of the original expressions.
  // Simply using floating point operations and random floats for symbols would be easiest. 
  // However, when we use floating point operations it is necessary to handle the inaccuracies 
  // in the last bits (e.g. for division, log and exp)
  // 
  // Collisions are OK as long as they occur seldom. 
  // It is probably be very unlikely to miss the optimal solution because of a collision. 
  //
  // We only need to identify equivalent structures. The numeric constants are irrelevant.
  // Therefore, equivalent structures with different numeric constants map to the same hash code.
  
  public static class ExprHash {
    const int MaxStackSize = 100;
    const int MaxVariables = 1000;
    private static double[] varSymbValues; 
    static ExprHash() {
      var rand = new MersenneTwister();
      varSymbValues = new double[MaxVariables];
      for(int i = 0;i < MaxVariables;i++) {
        varSymbValues[i] = rand.NextDouble();
      }
    }

    public static ulong GetHash(byte[] code, int nParams) {
      var bits = (ulong)BitConverter.DoubleToInt64Bits(Exec(code, nParams));
      // clear last five bits (insignificant?)
      bits = bits & 0xFFFFFFFFFFFFFFE0;
      return bits;
    }

    private static double Exec(byte[] code, int nParams) {
      var stack = new double[MaxStackSize];
      int topOfStack = -1;
      int pc = 0;
      int nextParamIdx = -1;
      OpCodes op;
      short arg;
      while (true) {
        ReadNext(code, ref pc, out op, out arg);
        switch (op) {
          case OpCodes.Nop: break;
          case OpCodes.LoadConst0: {
              ++topOfStack;
              stack[topOfStack] = 0.0;
              break;
            }
          case OpCodes.LoadConst1: {
              ++topOfStack;
              stack[topOfStack] = 1.0;
              break;
            }
          case OpCodes.LoadParamN: {
              ++topOfStack;
              stack[topOfStack] = 1.0;
              break;
            }
          case OpCodes.LoadVar: {
              ++topOfStack;
              stack[topOfStack] = varSymbValues[arg];
              break;
            }
          case OpCodes.Add: {
              var t1 = stack[topOfStack - 1];
              var t2 = stack[topOfStack];
              topOfStack--;
              stack[topOfStack] = t1 + t2;
              break;
            }
          case OpCodes.Mul: {
              var t1 = stack[topOfStack - 1];
              var t2 = stack[topOfStack];
              topOfStack--;
              stack[topOfStack] = t1 * t2;
              break;
            }
          case OpCodes.Log: {
              var v1 = stack[topOfStack];
              stack[topOfStack] = Math.Log(v1);
              break;
            }
          case OpCodes.Exp: {
              var v1 = stack[topOfStack];
              stack[topOfStack] = Math.Exp(v1);
              break;
            }
          case OpCodes.Inv: {
              var v1 = stack[topOfStack];
              stack[topOfStack] = 1.0 / v1;
              break;
            }
          case OpCodes.Exit:
            Contract.Assert(topOfStack == 0);
            return stack[topOfStack];
        }
      }
    }

    private static void ReadNext(byte[] code, ref int pc, out OpCodes op, out short s) {
      op = (OpCodes)Enum.ToObject(typeof(OpCodes), code[pc++]);
      s = 0;
      if (op == OpCodes.LoadVar) {
        s = (short)((code[pc] << 8) | code[pc + 1]);
        pc += 2;
      }
    }
  }
}
